//rewritten by Zheng Gengbin
// deal with shadow area
// add immediate SHIFT

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include "compilerApi.h"
#include "strategy.h"

#include "anls_att.h"

Shadow_def::Shadow_def(Org tname)
{
  tag=tname;
  for (int i=0; i<MAX_SHADOW_DIM; i++){
    l[i]=0; 
    u[i]=0;
  }
}

Org Shadow_def::getTag()
{
  return tag;
}

void Shadow_def::setLower(int dim, int size)
{
  l[dim]=size;
}
  
void Shadow_def::setUpper(int dim, int size)
{
  u[dim]=size;
}

int Shadow_def::getLower(int dim)
{
  return l[dim];
}

int Shadow_def::getUpper(int dim)
{
  return u[dim];
}

void  Shadow_def::setScope(SgStatement *s)
{
  scope=s;
}

SgStatement *  Shadow_def::getScope()
{
  return scope;
}

void Shadow_def::show(void)
{

  printf ("tag: %d\n", tag);
  printf("lsize : ");
  for(int i=0; i<MAX_SHADOW_DIM;i++)
    printf("%3d", l[i]);
  printf("\n");

  printf("usize : ");
  for(int j=0; j<MAX_SHADOW_DIM;j++)
    printf("%3d", u[j]);
  printf("\n");

  printf("\n");
}

//--------------

Shadow_def * get_defined_shadow(SgSymbol *sym)
{
    for  (int i=0; i<sym->numberOfAttributes(Shad_Def) ; i++) {
	Shadow_def *sdef = (Shadow_def *)sym->attributeValue(i,Shad_Def);
	if (sdef->getTag() == DEFINED) return sdef;
    }
    return NULL;
}

int shift_is_ok(SgStatement *s, SgSymbol *arraysymb)
{
  int can_shift = 1;
// is dummy
/*
// RUNTIME can already analyse the ghostarea now
  if (is_dummy(s,arraysymb) == 1) {
	if (get_defined_shadow(arraysymb) == NULL)
  		can_shift=0;
  }
*/
  return can_shift; 
}

// create a Shadow_def(GNEEDED) for array
void decide_global_shadow(SgStatement *s)
{ 

  for (int i=0; i<s->numberOfAttributes(Align_Def); i++) {
    Alignment_def *adef = (Alignment_def *)s->attributeValue(i, Align_Def);

    SgSymbol * array = s->getSymbolInCurrentScope(adef->getfromsymbol()->identifier());

    if (array->numberOfAttributes(Shad_Def)) {

      Shadow_def * defined_sdef = get_defined_shadow(array);

      Shadow_def * dsdef = new Shadow_def (GNEEDED);
      
      for (int numOfDim=0; numOfDim < MAX_ALIGN_DIM; numOfDim++) {
	int lsize=0;
	int usize=0;
	
	  // this strategy created GNEEDED by comparing all the DETECTED shadows
	  // or we create GNEEDED only by defined
	if (defined_sdef) {
	      lsize = defined_sdef->getLower(numOfDim);
	      usize = defined_sdef->getUpper(numOfDim);
	}
	else
	for ( int numS=0; numS<array->numberOfAttributes(Shad_Def); numS++) 
	{
	  Shadow_def * sdef = (Shadow_def *)array->attributeValue(numS, Shad_Def);
	  
	  //sdef->show();
	  
	  if (sdef->getTag() == DETECTED) {
	    if (sdef->getLower(numOfDim) > lsize)
	      lsize = sdef->getLower(numOfDim);
	    if (sdef->getUpper(numOfDim) > usize)
	      usize = sdef->getUpper(numOfDim);
	  }
	  
	}
	
	dsdef->setLower(numOfDim, lsize);
	dsdef->setUpper(numOfDim, usize);
	
      }
      
//      printf("here is global\n");
//      dsdef->show();
      
      array->addAttribute(Shad_Def,(void *) dsdef, sizeof(Shadow_def));
    }
  }
}

//---------------------------------;


void get_shift_array(SgStatement *s)
{
  SgStatement * body;
  body = s->childList1(0);

  Forall_header_def * fhdef = (Forall_header_def *)s->attributeValue(0,FHDER_Def);

  for (int i=0; i<body->numberOfAttributes(Forall_Ref); i++) {
    forall_reference * fref = (forall_reference *)body->attributeValue(i, Forall_Ref);

    if (fref->getCommunication()==SHIFT && !fref->isIShift() ) {
      int found =0;

      for ( int j=0; j<body->numberOfAttributes(SHIFTARRAY_Def); j++) {
	SgSymbol *array = (SgSymbol *)body->attributeValue(j, SHIFTARRAY_Def);
	
	if (array==fref->getexpression()->symbol()) {
	  found =1;
	  break;
	};
      };

      if (!found) {
	body->addAttribute(SHIFTARRAY_Def, fref->getexpression()->symbol(), sizeof(SgSymbol)); 
      }
      
    }
  }
}


// create a Shadow_def(LNEEDED) for array
void decide_local_shadow(SgStatement * s)
{

  SgStatement * body;
  body = s->childList1(0);

  for (int i=0; i<body->numberOfAttributes(SHIFTARRAY_Def); i++) {
    SgSymbol *array = (SgSymbol *)body->attributeValue(i, SHIFTARRAY_Def);

    if (array->numberOfAttributes(Shad_Def)) {

      Shadow_def * dsdef = new Shadow_def (LNEEDED);
      
      for (int numOfDim=0; numOfDim < MAX_ALIGN_DIM; numOfDim++) {
	int lsize=0;
	int usize=0;
	
	for ( int numS=0; numS<array->numberOfAttributes(Shad_Def); numS++) {
	  Shadow_def * sdef = (Shadow_def *)array->attributeValue(numS, Shad_Def);
	  if (sdef->getTag()==DETECTED && sdef->getScope()==s) {
	    
	    if (sdef->getLower(numOfDim) > lsize)
	      lsize = sdef->getLower(numOfDim);
	    if (sdef->getUpper(numOfDim) > usize)
	      usize = sdef->getUpper(numOfDim);
	  }
	  
	}
      
	dsdef->setLower(numOfDim, lsize);
	dsdef->setUpper(numOfDim, usize);      
      }
      
      //dsdef->show();
      dsdef->setScope(s);
      
      array->addAttribute(Shad_Def,(void *) dsdef, sizeof(Shadow_def));
    } // end of if
  } // end of for
}

/***************************************************************************/
//		Immediately shift
/***************************************************************************/

void define_unknow_ishift_shadow(SgStatement *s, int counter)
{
      SgStatement * body = s->childList1(0);
      forall_reference *fref = (forall_reference *)body->attributeValue(counter, Forall_Ref);
      SgSymbol *array = fref->getexpression()->symbol();
      Shadow_def * sdef = fref->getShadow();

      SgArrayType * arrayType=(SgArrayType *)array->type();
      int dim = arrayType->dimension();

      SgSymbol * amount=s->getSymbolInCurrentScope(mixstring, "amount", NULL,counter);

      int dimcounter = 0;
      for(int j=0; j<dim; j++){

	if (fref->indexNumberOnDim(j) == -1) continue;

	dimcounter++;

  	SgArrayRefExp *amountExp = new SgArrayRefExp(*amount);
	SgExpression *indexExp = new SgValueExp(j+1);
  	amountExp->addSubscript(*indexExp);
  	SgExpression *amountExp1 = &( SgUMinusOp(*amountExp->copyPtr()));

	  // name gtl_a?
	SgSymbol * gtl=s->getSymbolInCurrentScope(mixstring, "gtl", array->identifier(),counter);
	SgSymbol * gtu=s->getSymbolInCurrentScope(mixstring, "gtu", array->identifier(),counter);
      
	SgValueExp * dimExp=new SgValueExp(dimcounter);
	SgExpression * range = new SgExpression(RANGE_OP, NULL, dimExp, NULL);
	SgExpression * ddot = new SgExpression(DDOT, range, NULL, NULL);
	SgExpression * index = new SgExpression(EXPR_LIST, ddot, NULL, NULL);

	SgExpression * gtlarray = new SgExpression(ARRAY_REF, index, NULL, gtl);
	SgExpression * gtuarray = new SgExpression(ARRAY_REF, index, NULL, gtu);

  	SgSymbol * maxsymb=s->getSymbolInCurrentScope("max");
  	SgFunctionRefExp * maxcall1 = new SgFunctionRefExp (*maxsymb);
  	maxcall1->addArg(*amountExp1);    
  	maxcall1->addArg(*new SgValueExp(0));
	SgAssignStmt * gtlAssign = new SgAssignStmt(*gtlarray, *maxcall1);
  	SgFunctionRefExp *maxcall2 = new SgFunctionRefExp (*maxsymb);
  	maxcall2->addArg(*amountExp);    
  	maxcall2->addArg(*new SgValueExp(0));
	SgAssignStmt * gtuAssign = new SgAssignStmt(*gtuarray, *maxcall2);

	s->insertStmtBefore(*gtlAssign, *s->controlParent());
	s->insertStmtBefore(*gtuAssign, *s->controlParent());
      }
}

void define_ishift_shadow(SgStatement *s, int counter)
{
      SgStatement * body = s->childList1(0);
      forall_reference *fref = (forall_reference *)body->attributeValue(counter, Forall_Ref);
      SgSymbol *array = fref->getexpression()->symbol();
      Shadow_def * sdef = fref->getShadow();

	// unknow communication.
      if (fref->getCommunication() == UNKNOW) {
	define_unknow_ishift_shadow(s,counter);
	return;
      }

      SgArrayType * arrayType=(SgArrayType *)array->type();
      int dim = arrayType->dimension();

      int dimcounter = 0;
      for(int j=0; j<dim; j++){

	if (fref->indexNumberOnDim(j) == -1) continue;

	dimcounter++;
	SgValueExp * dimExp=new SgValueExp(dimcounter);

	SgValueExp * lowerExp= new SgValueExp(sdef->getLower(j));
	SgValueExp * upperExp= new SgValueExp(sdef->getUpper(j));

	SgSymbol * gtl=s->getSymbolInCurrentScope(mixstring, "gtl", array->identifier(), counter);
	SgSymbol * gtu=s->getSymbolInCurrentScope(mixstring, "gtu", array->identifier(), counter);
      
	SgExpression * range = new SgExpression(RANGE_OP, NULL, dimExp, NULL);
	SgExpression * ddot = new SgExpression(DDOT, range, NULL, NULL);
	SgExpression * index = new SgExpression(EXPR_LIST, ddot, NULL, NULL);

	SgExpression * gtlarray = new SgExpression(ARRAY_REF, index, NULL, gtl);
	SgExpression * gtuarray = new SgExpression(ARRAY_REF, index, NULL, gtu);

	SgAssignStmt * gtlAssign = new SgAssignStmt(*gtlarray, *lowerExp);
	SgAssignStmt * gtuAssign = new SgAssignStmt(*gtuarray, *upperExp);

	s->insertStmtBefore(*gtlAssign, *s->controlParent());
	s->insertStmtBefore(*gtuAssign, *s->controlParent());

      }
}

void add_ishift_write_halo(SgStatement *s, int counter)
{
      SgStatement * body = s->childList1(0);
      forall_reference *fref = (forall_reference *)body->attributeValue(counter, Forall_Ref);
      SgSymbol *array = fref->getexpression()->symbol();

      SgSymbol * dad=s->getSymbolInCurrentScope(mixstring, "tdd", NULL,counter);  
      SgSymbol *gtl=s->getSymbolInCurrentScope(mixstring, "gtl", array->identifier(),counter);
      SgSymbol *gtu=s->getSymbolInCurrentScope(mixstring, "gtu", array->identifier(),counter);
      SgStatement *func = make_write_halo(s, dad,gtl, gtu);
      s->insertStmtBefore(*func, *s->controlParent());
}


/**************************************************************************/

void define_shadow_amount(SgStatement * s, int counter)
{
  SgStatement * body;
  body = s->childList1(0);

  SgSymbol * array = (SgSymbol *) ((body->getAttribute(counter, SHIFTARRAY_Def))->getAttributeData());

  //printf("define halo for %s\n", array->identifier());

  for (int i=0; i<array->numberOfAttributes(Shad_Def); i++) {
    Shadow_def * sdef = (Shadow_def *)array->attributeValue(i, Shad_Def);
    if (sdef->getTag()==LNEEDED && sdef->getScope()==s) {
	  // create assignment to define shadow
      SgArrayType * arrayType=(SgArrayType *)array->type();
      int dim = arrayType->dimension();

      //printf("type : %d dim %d\n", arrayType->id(), dim);
      
      for(int j=0; j<dim; j++){

	SgValueExp * dimExp=new SgValueExp(j+1);

	SgValueExp * lowerExp= new SgValueExp(sdef->getLower(j));
	SgValueExp * upperExp= new SgValueExp(sdef->getUpper(j));

	SgSymbol * gtl=s->getSymbolInCurrentScope(mixstring, "gtl", array->identifier());
	SgSymbol * gtu=s->getSymbolInCurrentScope(mixstring, "gtu", array->identifier());
      
	SgExpression * range = new SgExpression(RANGE_OP, NULL, dimExp, NULL);
	SgExpression * ddot = new SgExpression(DDOT, range, NULL, NULL);
	SgExpression * index = new SgExpression(EXPR_LIST, ddot, NULL, NULL);

	SgExpression * gtlarray = new SgExpression(ARRAY_REF, index, NULL, gtl);
	SgExpression * gtuarray = new SgExpression(ARRAY_REF, index, NULL, gtu);

	SgAssignStmt * gtlAssign = new SgAssignStmt(*gtlarray, *lowerExp);
	SgAssignStmt * gtuAssign = new SgAssignStmt(*gtuarray, *upperExp);

	s->insertStmtBefore(*gtlAssign, *s->controlParent());
	s->insertStmtBefore(*gtuAssign, *s->controlParent());

      }
      break;
    }
  }
}

void add_write_halo(SgStatement * s, int counter)
{
  SgStatement * body;
  body = s->childList1(0);

  SgSymbol * array = (SgSymbol *) ((body->getAttribute(counter, SHIFTARRAY_Def))->getAttributeData());

  //printf("write halo for %s\n", array->identifier());

  SgSymbol * dad=s->getSymbolInCurrentScope(mixstring, "dad", array->identifier());  
  
  SgSymbol * gtl=s->getSymbolInCurrentScope(mixstring, "gtl", array->identifier());

  SgSymbol * gtu=s->getSymbolInCurrentScope(mixstring, "gtu", array->identifier());
  
  SgStatement *func = make_write_halo(s, dad,gtl, gtu);

  s->insertStmtBefore(*func, *s->controlParent());
}


void write_halo(SgStatement *s)
{
  SgStatement * body;
  body = s->childList1(0);

  for (int i=0; i<body->numberOfAttributes(SHIFTARRAY_Def); i++) {
    SgSymbol *array = (SgSymbol *)body->attributeValue(i, SHIFTARRAY_Def);

    if (array->numberOfAttributes(Shad_Def)) {

      define_shadow_amount(s, i);
      add_write_halo(s, i);

    }
  }
}

/*****************************************************************************
			UNKNOW SHIFT
    we assume shift will happen here in unknow case.
*****************************************************************************/

void get_unknow_shift_array(SgStatement *s)
{
  SgStatement * body;
  body = s->childList1(0);

  Forall_header_def * fhdef = (Forall_header_def *) s->getAttribute(0, FHDER_Def)->getAttributeData();

  for (int i=0; i<body->numberOfAttributes(Forall_Ref); i++) {
    forall_reference * fref = (forall_reference *)body->attributeValue(i, Forall_Ref);
    SgSymbol *arraysymb = fref->getexpression()->symbol();
    int shift_ok=shift_is_ok(s, arraysymb);
    if (!shift_ok) continue;

    if (fref->getCommunication()==UNKNOW && !fref->isIShift() ) {
      int found =0;

      for ( int j=0; j<body->numberOfAttributes(UNKNOWSHIFT_Def); j++) {
	SgSymbol *array = (SgSymbol *)body->attributeValue(j, UNKNOWSHIFT_Def);
	
	if (array==arraysymb) {
	  found =1;
	  break;
	};
      };

      if (!found) {
	body->addAttribute(UNKNOWSHIFT_Def, arraysymb, sizeof(SgSymbol)); 
      }
      
    }
  }

  // assign initial value
  for ( int j=0; j<body->numberOfAttributes(UNKNOWSHIFT_Def); j++) {
	SgSymbol *arraysymb = (SgSymbol *)body->attributeValue(j, UNKNOWSHIFT_Def);
	SgArrayType *type = isSgArrayType(arraysymb->type());
	ASSERT(type);
	// assign init value comm_A:0
	{
  	SgSymbol * comm=s->getSymbolInCurrentScope(mixstring, "comshift", arraysymb->identifier());
	SgExpression *commExp = new SgVarRefExp(comm);
	SgStatement *assign = new SgAssignStmt(*commExp, *new SgBoolValueExp(0));
  	s->insertStmtBefore(*assign, *s->controlParent());
	}
	// assign init value of gtl and gtu
	{
  	SgSymbol * gtl=s->getSymbolInCurrentScope(mixstring, "gtl", arraysymb->identifier());
  	SgArrayRefExp *gtlExp = new SgArrayRefExp(*gtl);
  	SgSymbol * gtu=s->getSymbolInCurrentScope(mixstring, "gtu", arraysymb->identifier());
  	SgArrayRefExp *gtuExp = new SgArrayRefExp(*gtu);
  	SgSymbol * indexsymb=s->getSymbolInCurrentScope(mixstring, "tmp","sdx");
  	SgExpression *indexExp = new SgVarRefExp(indexsymb);
  	gtlExp->addSubscript(*indexExp);
  	gtuExp->addSubscript(*indexExp);
	SgStatement *ass1 = new SgAssignStmt(*gtlExp, *new SgValueExp(0));
	SgStatement *ass2 = new SgAssignStmt(*gtuExp, *new SgValueExp(0));
	int dim = type->dimension();
  	SgStatement *dostmt = new SgForStmt(*indexsymb, *new SgValueExp(1), *new SgValueExp(dim), *ass1);
  	s->insertStmtBefore(*dostmt, *s->controlParent());
  	ass1->insertStmtAfter(*ass2, *ass1->controlParent());
	}
  }
}

// write_halo for unknow mode
void unknow_write_halo(SgStatement *s)
{
  SgStatement * body;
  body = s->childList1(0);

  for (int i=0; i<body->numberOfAttributes(UNKNOWSHIFT_Def); i++) {
    SgSymbol *array = (SgSymbol *)body->attributeValue(i, UNKNOWSHIFT_Def);

  	SgSymbol * dad=s->getSymbolInCurrentScope(mixstring, "dad", array->identifier());  
  	SgSymbol * gtl=s->getSymbolInCurrentScope(mixstring, "gtl", array->identifier());
  	SgSymbol * gtu=s->getSymbolInCurrentScope(mixstring, "gtu", array->identifier());
	SgStatement *halo = make_write_halo(s, dad,gtl, gtu);

  	SgSymbol * amount=s->getSymbolInCurrentScope(mixstring, "comshift", array->identifier());
   	SgExpression *amountExp = new SgVarRefExp(amount);
//  	SgExpression *conExp = &(SgEqOp(*amountExp, *new SgValueExp(1)));
	SgExpression *conExp = amountExp;
	SgStatement *ifStmt = new SgLogIfStmt(*conExp , *halo);
  	s->insertStmtBefore(*ifStmt, *s->controlParent());
  }
}

void check_ghost(SgStatement * s)
{

  get_shift_array(s);

  // add by Zheng Gengbin
  if ( opt_strategy & RUNTIMECOMM )
  {
  // collect shift array and assign initial values
  get_unknow_shift_array(s);
  }

  decide_local_shadow(s);
}

void add_ghost(SgStatement * s)
{
  write_halo(s);

  // add by Zheng Gengbin
  if ( opt_strategy & RUNTIMECOMM )
  {
  unknow_write_halo(s);
  }
}


